/*
Copyright (c) 2000-2011 by Nicolas Devillard.
MIT License

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/*-------------------------------------------------------------------------*/
/**
   @file    iniparser.c
   @author  N. Devillard
   @brief   Parser for ini files.
*/

/*---------------------------- Includes ------------------------------------*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <ctype.h>

#include <atalk/iniparser.h>
#include <atalk/logger.h>

/*---------------------------- Defines -------------------------------------*/
#define ASCIILINESZ (1024)
#define INI_INVALID_KEY ((char *)-1)

/*---------------------------------------------------------------------------
                        Private to this module
 ---------------------------------------------------------------------------*/
/**
 * This enum stores the status for each parsed line (internal use only).
 */
typedef enum _line_status_ {
    LINE_UNPROCESSED,
    LINE_ERROR,
    LINE_EMPTY,
    LINE_COMMENT,
    LINE_SECTION,
    LINE_VALUE
} line_status;

/*-------------------------------------------------------------------------*/
/**
  @brief    Duplicate a string
  @param    s String to duplicate
  @return   Pointer to a newly allocated string, to be freed with free()

  This is a replacement for strdup(). This implementation is provided
  for systems that do not have it.
 */
/*--------------------------------------------------------------------------*/
static char *xstrdup(const char *s)
{
    char *t;
    size_t len;
    if (!s)
        return NULL;

    len = strlen(s) + 1;
    t = (char *)malloc(len);
    if (t) {
        memcpy(t, s, len);
    }
    return t;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Remove blanks at the beginning and the end of a string.
  @param    str  String to parse and alter.
  @return   unsigned New size of the string.
 */
/*--------------------------------------------------------------------------*/
static unsigned strstrip(char *s)
{
    char *last = NULL;
    char *dest = s;

    if (s == NULL)
        return 0;

    last = s + strlen(s);
    while (isspace((int)*s) && *s)
        s++;
    while (last > s) {
        if (!isspace((int)*(last - 1)))
            break;
        last--;
    }
    *last = (char)0;

    memmove(dest, s, last - s + 1);
    return last - s;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get number of sections in a dictionary
  @param    d   Dictionary to examine
  @return   int Number of sections found in dictionary

  This function returns the number of sections found in a dictionary.
  The test to recognize sections is done on the string stored in the
  dictionary: a section name is given as "section" whereas a key is
  stored as "section:key", thus the test looks for entries that do not
  contain a colon.

  This clearly fails in the case a section name contains a colon, but
  this should simply be avoided.

  This function returns -1 in case of error.
 */
/*--------------------------------------------------------------------------*/
int atalk_iniparser_getnsec(const dictionary *d)
{
    int i;
    int nsec;

    if (d == NULL)
        return -1;
    nsec = 0;
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL)
            continue;
        if (strchr(d->key[i], ':') == NULL) {
            nsec++;
        }
    }
    return nsec;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get name for section n in a dictionary.
  @param    d   Dictionary to examine
  @param    n   Section number (from 0 to nsec-1).
  @return   Pointer to char string

  This function locates the n-th section in a dictionary and returns
  its name as a pointer to a string statically allocated inside the
  dictionary. Do not free or modify the returned string!

  This function returns NULL in case of error.
 */
/*--------------------------------------------------------------------------*/
const char *atalk_iniparser_getsecname(const dictionary *d, int n)
{
    int i;
    int foundsec;

    if (d == NULL || n < 0)
        return NULL;
    foundsec = 0;
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL)
            continue;
        if (strchr(d->key[i], ':') == NULL) {
            foundsec++;
            if (foundsec > n)
                break;
        }
    }
    if (foundsec <= n) {
        return NULL;
    }
    return d->key[i];
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Dump a dictionary to an opened file pointer.
  @param    d   Dictionary to dump.
  @param    f   Opened file pointer to dump to.
  @return   void

  This function prints out the contents of a dictionary, one element by
  line, onto the provided file pointer. It is OK to specify @c stderr
  or @c stdout as output files. This function is meant for debugging
  purposes mostly.
 */
/*--------------------------------------------------------------------------*/
void atalk_iniparser_dump(const dictionary *d, FILE *f)
{
    int i;

    if (d == NULL || f == NULL)
        return;
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL)
            continue;
        if (d->val[i] != NULL) {
            fprintf(f, "[%s]=[%s]\n", d->key[i], d->val[i]);
        } else {
            fprintf(f, "[%s]=UNDEF\n", d->key[i]);
        }
    }
    return;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Save a dictionary section to a loadable ini file
  @param    d   Dictionary to dump
  @param    s   Section name of dictionary to dump
  @param    f   Opened file pointer to dump to
  @return   void

  This function dumps a given section of a given dictionary into a loadable ini
  file.  It is Ok to specify @c stderr or @c stdout as output files.
 */
/*--------------------------------------------------------------------------*/
void atalk_iniparser_dumpsection_ini(const dictionary *d, const char *s, FILE *f)
{
    int j;
    char keym[ASCIILINESZ + 1];
    int seclen;

    if (d == NULL || f == NULL)
        return;
    if (!atalk_iniparser_find_entry(d, s))
        return;

    seclen = (int)strlen(s);
    fprintf(f, "\n[%s]\n", s);
    snprintf(keym, sizeof(keym), "%s:", s);
    for (j = 0; j < d->size; j++) {
        if (d->key[j] == NULL)
            continue;
        if (!strncmp(d->key[j], keym, seclen + 1)) {
            fprintf(f,
                    "%-30s = %s\n",
                    d->key[j] + seclen + 1,
                    d->val[j] ? d->val[j] : "");
        }
    }
    fprintf(f, "\n");
    return;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Save a dictionary to a loadable ini file
  @param    d   Dictionary to dump
  @param    f   Opened file pointer to dump to
  @return   void

  This function dumps a given dictionary into a loadable ini file.
  It is Ok to specify @c stderr or @c stdout as output files.
 */
/*--------------------------------------------------------------------------*/
void atalk_iniparser_dump_ini(const dictionary *d, FILE *f)
{
    int i;
    int nsec;
    const char *secname;

    if (d == NULL || f == NULL)
        return;

    nsec = atalk_iniparser_getnsec(d);
    if (nsec < 1) {
        /* No section in file: dump all keys as they are */
        for (i = 0; i < d->size; i++) {
            if (d->key[i] == NULL)
                continue;
            fprintf(f, "%s = %s\n", d->key[i], d->val[i]);
        }
        return;
    }
    for (i = 0; i < nsec; i++) {
        secname = atalk_iniparser_getsecname(d, i);
        atalk_iniparser_dumpsection_ini(d, secname, f);
    }
    fprintf(f, "\n");
    return;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key
  @param    d       Dictionary to search
  @param    section Section to search
  @param    key     Key string to look for
  @param    def     Default value to return if key not found.
  @return   pointer to statically allocated character string

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the pointer passed as 'def' is returned.
  The returned char pointer is pointing to a string allocated in
  the dictionary, do not free or modify it.
 */
/*--------------------------------------------------------------------------*/
const char *atalk_iniparser_getstring(const dictionary *d, const char *section, const char *key, const char *def)
{
    const char *sval;

    if (d == NULL || key == NULL)
        return def;

    sval = atalkdict_get(d, section, key, def);
    return sval;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key
  @param    d       Dictionary to search
  @param    section Section to search
  @param    key     Key string to look for
  @param    def     Default value to return if key not found.
  @return   pointer to statically allocated character string

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the pointer passed as 'def' is returned.
  The returned char pointer a strdup'ed allocated string, so the caller must free it.
 */
/*--------------------------------------------------------------------------*/
char *atalk_iniparser_getstrdup(const dictionary *d, const char *section, const char *key, const char *def)
{
    const char *sval;

    if (d == NULL || key == NULL)
        return NULL;

    if ((sval = atalkdict_get(d, section, key, def)))
        return strdup(sval);
    return NULL;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key, convert to an long int
  @param    d Dictionary to search
  @param    section  Section to search
  @param    key Key string to look for
  @param    notfound Value to return in case of error
  @return   long integer

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the notfound value is returned.

  Supported values for integers include the usual C notation
  so decimal, octal (starting with 0) and hexadecimal (starting with 0x)
  are supported. Examples:

  "42"      ->  42
  "042"     ->  34 (octal -> decimal)
  "0x42"    ->  66 (hexa  -> decimal)

  Warning: the conversion may overflow in various ways. Conversion is
  totally outsourced to strtol(), see the associated man page for overflow
  handling.

  Credits: Thanks to A. Becker for suggesting strtol()
 */
/*--------------------------------------------------------------------------*/
long int atalk_iniparser_getlongint(const dictionary *d, const char *section, const char *key, long int notfound)
{
    const char *str;

    str = atalk_iniparser_getstring(d, section, key, INI_INVALID_KEY);
    if (str == INI_INVALID_KEY)
        return notfound;
    return strtol(str, NULL, 0);
}


/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key, convert to an int
  @param    d Dictionary to search
  @param    section  Section to search
  @param    key Key string to look for
  @param    notfound Value to return in case of error
  @return   integer

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the notfound value is returned.

  Supported values for integers include the usual C notation
  so decimal, octal (starting with 0) and hexadecimal (starting with 0x)
  are supported. Examples:

  "42"      ->  42
  "042"     ->  34 (octal -> decimal)
  "0x42"    ->  66 (hexa  -> decimal)

  Warning: the conversion may overflow in various ways. Conversion is
  totally outsourced to strtol(), see the associated man page for overflow
  handling.

  Credits: Thanks to A. Becker for suggesting strtol()
 */
/*--------------------------------------------------------------------------*/
int atalk_iniparser_getint(const dictionary *d, const char *section, const char *key, int notfound)
{
    return (int)atalk_iniparser_getlongint(d, section, key, notfound);
}

/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key, convert to a double
  @param    d        Dictionary to search
  @param    section  Section to search
  @param    key      Key string to look for
  @param    notfound Value to return in case of error
  @return   double

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the notfound value is returned.
 */
/*--------------------------------------------------------------------------*/
double atalk_iniparser_getdouble(const dictionary *d, const char *section, const char *key, double notfound)
{
    const char *str;

    str = atalk_iniparser_getstring(d, section, key, INI_INVALID_KEY);
    if (str == INI_INVALID_KEY)
        return notfound;
    return atof(str);
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Get the string associated to a key, convert to a boolean
  @param    d        Dictionary to search
  @param    section  Section to search
  @param    key      Key string to look for
  @param    notfound Value to return in case of error
  @return   integer

  This function queries a dictionary for a key. A key as read from an
  ini file is given as "section:key". If the key cannot be found,
  the notfound value is returned.

  A true boolean is found if one of the following is matched:

  - A string starting with 'y'
  - A string starting with 'Y'
  - A string starting with 't'
  - A string starting with 'T'
  - A string starting with '1'

  A false boolean is found if one of the following is matched:

  - A string starting with 'n'
  - A string starting with 'N'
  - A string starting with 'f'
  - A string starting with 'F'
  - A string starting with '0'

  The notfound value returned if no boolean is identified, does not
  necessarily have to be 0 or 1.
 */
/*--------------------------------------------------------------------------*/
int atalk_iniparser_getboolean(const dictionary *d, const char *section, const char *key, int notfound)
{
    const char *c;
    int ret;

    c = atalk_iniparser_getstring(d, section, key, INI_INVALID_KEY);
    if (c == INI_INVALID_KEY)
        return notfound;
    if (c[0] == 'y' || c[0] == 'Y' || c[0] == '1' || c[0] == 't' || c[0] == 'T') {
        ret = 1;
    } else if (c[0] == 'n' || c[0] == 'N' || c[0] == '0' || c[0] == 'f' || c[0] == 'F') {
        ret = 0;
    } else {
        ret = notfound;
    }
    return ret;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Finds out if a given entry exists in a dictionary
  @param    ini     Dictionary to search
  @param    entry   Name of the entry to look for
  @return   integer 1 if entry exists, 0 otherwise

  Finds out if a given entry exists in the dictionary. Since sections
  are stored as keys with NULL associated values, this is the only way
  of querying for the presence of sections in a dictionary.
 */
/*--------------------------------------------------------------------------*/
int atalk_iniparser_find_entry(const dictionary *ini, const char *entry)
{
    int found = 0;
    if (atalk_iniparser_getstring(ini, entry, NULL, INI_INVALID_KEY) != INI_INVALID_KEY) {
        found = 1;
    }
    return found;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Set an entry in a dictionary.
  @param    ini     Dictionary to modify.
  @param    section Entry to modify (entry section)
  @param    key     Entry to modify (entry key)
  @param    val     New value to associate to the entry.
  @return   int 0 if Ok, -1 otherwise.

  If the given entry can be found in the dictionary, it is modified to
  contain the provided value. If it cannot be found, -1 is returned.
  It is Ok to set val to NULL.
 */
/*--------------------------------------------------------------------------*/
int atalk_iniparser_set(dictionary *ini, const char *section, const char *key, const char *val)
{
    return atalkdict_set(ini, section, key, val);
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Delete an entry in a dictionary
  @param    ini     Dictionary to modify
  @param    section Entry to delete (entry section)
  @param    key     Entry to delete (entry key)
  @return   void

  If the given entry can be found, it is deleted from the dictionary.
 */
/*--------------------------------------------------------------------------*/
void atalk_iniparser_unset(dictionary *ini, const char *section, const char *key)
{
    atalkdict_unset(ini, section, key);
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Load a single line from an INI file
  @param    input_line  Input line, may be concatenated multi-line input
  @param    section     Output space to store section
  @param    key         Output space to store key
  @param    value       Output space to store value
  @return   line_status value
 */
/*--------------------------------------------------------------------------*/
static line_status atalk_iniparser_line(
    const char *input_line,
    char *section,
    char *key,
    char *value)
{
    line_status sta;
    char *line = NULL;
    size_t len;

    line = xstrdup(input_line);
    len = strstrip(line);

    sta = LINE_UNPROCESSED;
    if (len < 1) {
        /* Empty line */
        sta = LINE_EMPTY;
    } else if (line[0] == '#' || line[0] == ';') {
        /* Comment line */
        sta = LINE_COMMENT;
    } else if (line[0] == '[' && line[len - 1] == ']') {
        /* Section name */
        sscanf(line, "[%[^]]", section);
        strstrip(section);
        sta = LINE_SECTION;
    } else if (sscanf(line, "%[^=] = '%[^\']'", key, value) == 2) {
        /* Usual key=value with quotes, with or without comments */
        strstrip(key);
        /* Don't strip spaces from values surrounded with quotes */
        sta = LINE_VALUE;
    } else if (sscanf(line, "%[^=] = %[^;#]", key, value) == 2) {
        /* Usual key=value without quotes, with or without comments */
        strstrip(key);
        strstrip(value);
        /*
         * sscanf cannot handle '' or "" as empty values
         * this is done here
         */
        if (!strcmp(value, "\"\"") || (!strcmp(value, "''"))) {
            value[0] = 0;
        }
        sta = LINE_VALUE;
    } else if (sscanf(line, "%[^=] = %[;#]", key, value) == 2 || sscanf(line, "%[^=] %[=]", key, value) == 2) {
        /*
         * Special cases:
         * key=
         * key=;
         * key=#
         */
        strstrip(key);
        value[0] = 0;
        sta = LINE_VALUE;
    } else {
        /* Generate syntax error */
        sta = LINE_ERROR;
    }

    free(line);
    return sta;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Parse an ini file and return an allocated dictionary object
  @param    ininame Name of the ini file to read.
  @return   Pointer to newly allocated dictionary

  This is the parser for ini files. This function is called, providing
  the name of the file to be read. It returns a dictionary object that
  should not be accessed directly, but through accessor functions
  instead.

  The returned dictionary must be freed using atalk_iniparser_freedict().
 */
/*--------------------------------------------------------------------------*/
dictionary *atalk_iniparser_load(const char *ininame)
{
    FILE *in, *include = NULL, *inifile;

    char line[ASCIILINESZ + 1];
    char section[ASCIILINESZ + 1];
    char key[ASCIILINESZ + 1];
    char val[ASCIILINESZ + 1];

    int last = 0;
    int len;
    int lineno = 0;
    int errs = 0;

    dictionary *dict;

    if ((inifile = fopen(ininame, "r")) == NULL) {
        LOG(log_error, logtype_default, "iniparser: cannot open \"%s\"", ininame);
        return NULL;
    }

    dict = atalkdict_new(0);
    if (!dict) {
        fclose(inifile);
        return NULL;
    }

    memset(line, 0, ASCIILINESZ);
    memset(section, 0, ASCIILINESZ);
    memset(key, 0, ASCIILINESZ);
    memset(val, 0, ASCIILINESZ);
    last = 0;

    in = inifile;
    while (fgets(line + last, ASCIILINESZ - last, in) != NULL) {
        lineno++;
        len = (int)strlen(line) - 1;
        if (len <= 0)
            continue;
        /* Safety check against buffer overflows */
        if (line[len] != '\n' && !feof(in)) {
            LOG(log_error, logtype_default, "iniparser: input line too long in %s (lineno: %d): %s");
            atalkdict_del(dict);
            fclose(in);
            return NULL;
        }
        /* Get rid of \n and spaces at end of line */
        while ((len >= 0) &&
               ((line[len] == '\n') || (isspace(line[len])))) {
            line[len] = 0;
            len--;
        }
        if (len < 0) { /* Line was entirely \n and/or spaces */
            len = 0;
        }
        /* Detect multi-line */
        if (line[len] == '\\') {
            /* Multi-line value */
            last = len;
            continue;
        } else {
            last = 0;
        }
        switch (atalk_iniparser_line(line, section, key, val)) {
            case LINE_EMPTY:
            case LINE_COMMENT:
                break;
            case LINE_SECTION:
                if (strchr(section, ':') != NULL)
                    LOG(log_error, logtype_default, "iniparser: syntax error \"%s\" section name must not contain \":\".", section);
                errs = atalkdict_set(dict, section, NULL, NULL);
                break;
            case LINE_VALUE:
                if (strcmp(key, "include") == 0) {
                    errs = atalkdict_set(dict, section, key, val);
                    if (errs < 0) {
                        LOG(log_error, logtype_default, "iniparser: memory allocation failure");
                    }
                    if ((include = fopen(val, "r")) == NULL) {
                        LOG(log_error, logtype_default, "iniparser: cannot open \"%s\"", val);
                        continue;
                    }
                    in = include;
                    continue;
                }
                errs = atalkdict_set(dict, section, key, val);
                break;
            case LINE_ERROR:
                LOG(log_error, logtype_default, "iniparser: syntax error in %s (lineno: %d): %s",
                    ininame, lineno, line);
                errs++;
                break;
            default:
                break;
        }
        memset(line, 0, ASCIILINESZ);
        last = 0;
        if (errs < 0) {
            LOG(log_error, logtype_default, "iniparser: memory allocation failure");
            break;
        }
    }
    if (errs) {
        atalkdict_del(dict);
        dict = NULL;
    }
    fclose(in);
    return dict;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Free all memory associated to an ini dictionary
  @param    d Dictionary to free
  @return   void

  Free all memory associated to an ini dictionary.
  It is mandatory to call this function before the dictionary object
  gets out of the current context.
 */
/*--------------------------------------------------------------------------*/
void atalk_iniparser_freedict(dictionary *d)
{
    atalkdict_del(d);
}
